import matplotlib.pyplot as plt
import numpy as np
import random
import math

n = 10

raw_x = [random.random() for x in range(n)]
raw_y = [2 + 3*i + random.gauss(0, 0.5) for i in raw_x]


def normalize(v):
    mean = sum(v)/len(v)
    shifted = [e-mean for e in v]
    std  = math.sqrt(sum([e*e for e in shifted])/len(v))
    return [e/std for e in shifted], mean, std

def linear_least_squares(raw_x, raw_y):
    n = len(raw_x)
    x, mean_x, std_x = normalize(raw_x)
    y, mean_y, std_y = normalize(raw_y)

    A = np.array([[1, i] for i in x])
    U, W_diag, V = np.linalg.svd(A, full_matrices=0)
    W_inv_diag = []
    for e in W_diag:
        if e==0.0:
            W_inv_diag.append(0)
        else:
            W_inv_diag.append(1/e)
    W_inv = np.diag(W_inv_diag)

    coefs = np.matrix(V) * np.matrix(W_inv) * np.matrix(np.transpose(U)) * np.transpose(np.matrix(y))

    c0 = coefs.item(0,0)*std_y - coefs.item(1,0) * mean_x * std_y/std_x + mean_y
    c1 = coefs.item(1,0)*std_y/std_x
    fit_error = []
    for i in range(len(W_diag)):
        fit_error.append(sum([(V.item(i,j)/W_diag.item(j))**2 for j in range(len(W_diag))]))
    return c0, c1, fit_error

def bootstrap(raw_x, raw_y):
    bs_x_samples=[]
    bs_y_samples=[]
    for j in range(len(raw_x)):
        i = random.randint(0, len(raw_x)-1)
        bs_x_samples.append(raw_x[i])
        bs_y_samples.append(raw_y[i])
    return linear_least_squares(bs_x_samples, bs_y_samples)

if __name__ == "__main__":
    #Calculate the coefficients, extracting the fit errors from the SVD
    plt.scatter(raw_x, raw_y)
    c0, c1, f = linear_least_squares(raw_x, raw_y)
    print "error in fit relative to datat:", f

    #Bootstrapping
    bs_c0s = []
    bs_c1s = []
    for i in range(100):
        bs_c0, bs_c1, bs_f = bootstrap(raw_x, raw_y)
        bs_c0s.append(bs_c0)
        bs_c1s.append(bs_c1)
    bs_c0_norm, bs_c0_mean, bs_c0_std = normalize(bs_c0s)
    bs_c1_norm, bs_c1_mean, bs_c1_std = normalize(bs_c1s)
    print "bootstrapping error in fit data", [bs_c0_std**2, bs_c1_std**2]

    #Ensemble
    e_c0s = []
    e_c1s = []
    for i in range(100):        
        e_raw_x = [random.random() for x in range(n)]
        e_raw_y = [2 + 3*i + random.gauss(0, 0.5) for i in e_raw_x]
        e_c0, e_c1, e_f = linear_least_squares(e_raw_x, e_raw_y)
        e_c0s.append(e_c0)
        e_c1s.append(e_c1)
    e_c0_norm, e_c0_mean, e_c0_std = normalize(e_c0s)
    e_c1_norm, e_c1_mean, e_c1_std = normalize(e_c1s)
    print "ensemble error in fit data", [e_c0_std**2, e_c1_std**2]

    plt.plot([0,1], [c0, c1+c0])
    plt.show()

